import { Component } from 'react'
import Taro from '@tarojs/taro'
import { View, Text, Image } from '@tarojs/components'

import { postPuzzles } from '../../api/puzzle'
import getBarHeight from '../../utils/getBarHeight'
import image from '../../assets/back.png'

import './index.scss'

const gameStartTime = '2021年7月12日 0:00:00'
const gameEndTime = '2021年7月15日 0:00:00'

// 游戏规则
const gameIntroduction = (puzzleStep, currentTime, restTime) => (
  [
    {
      id: 13,
      title: '游戏有效期',
      content: [
        `${gameStartTime} - ${gameEndTime}`
      ]
    },
    {
      id: 14,
      title: '游戏奖励',
      content: [
        `一等奖：在 ${puzzleStep} 步内完成，时间在 ${currentTime - restTime * 2} s 内；`,
        `二等奖：在任意步数内完成，时间在 ${currentTime - restTime} s 内；`,
        `三等奖：在任意步数 步内完成，时间在 ${currentTime} s 内；`
      ]
    },
    {
      id: 15,
      title: '游戏兑换',
      content: [
        '请查看我的积分里兑换相应产品；'
      ]
    }
  ]
)

const baseUrl = process.env.BASE_URL
export default class Puzzle extends Component {
  constructor(props) {
    super(props)
    this.state = {
      navigationBarStyle: {
        height: 88 + 'rpx'
      },
      blocks: [], // 实时拼图位置
      constanceBlocks: [], // 完整拼图
      whiteBlockIdex: -1, // 当前空白目标的序号
      chooseIndex: 8, // 用户所选方块变为空白
      // 初始数据（3*3、4*4）
      initData: {},

      // 条件判断
      nowSteps: 0, // 完成步数判断
      currentTime: 0, // 倒计时
      restTime: 5, // 二等奖 （游戏完成后剩余多长时间）

      // 倒计时
      isStart: false, // 倒计时开始
      timer: null, // 倒计时计时器
      firstStart: true, // 首次开始

      // 可控元素
      puzzleLevel: 0,
      puzzleStep: 16 // 乱序步数
      // puzzleTimes: 10 // 游戏次数
    }
  }

  componentWillMount() {
    this.init()
    this.setState({
      navigationBarStyle: getBarHeight()
    })
  }

  // 是否是空格方块周围相邻
  isNeighbour(index) { // 是否为相邻
    const { whiteBlockIdex, initData } = this.state
    const { neighbWhiteBlock } = initData
    console.log(whiteBlockIdex, neighbWhiteBlock[String(whiteBlockIdex)], index)
    return neighbWhiteBlock[String(whiteBlockIdex)].includes(index)
  }

  // 游戏是否完成
  isRecover(blocks, timeOver = false) {
    let puzzleIdx = 0
    // 游戏是否完成 -> 计算拼图块块是否对应各个序号
    for (let i = 0, len = blocks.length; i < len; i++) {
      const item = blocks[i]
      if (item.idx === i) {
        puzzleIdx++
      } else {
        // 游戏结束后触发
        if (timeOver) {
          this.setState({
            isStart: false
          })
          return this.postPuzzle('fail').then((times) => {
            if (times === undefined) return
            Taro.showToast({
              title: `谢谢参与，游戏次数剩余 ${times} 次`,
              icon: 'none'
            })
          })
        }
      }
    }
    // 每次移动是触发
    const { initData, restTime } = this.state
    const { rowNum } = initData
    if (puzzleIdx === rowNum * rowNum) { // 通关
      const { currentTime, timer, nowSteps, puzzleStep } = this.state
      let prize = ''
      if (currentTime >= restTime) {
        if (nowSteps <= puzzleStep && currentTime >= restTime * 2) {
          prize = '一'
        } else {
          prize = '二'
        }
      } else {
        prize = '三'
      }
      this.postPuzzle('success').then((times) => {
        if (times === undefined) return
        return Taro.showToast({
          title: `${prize}等奖，游戏次数剩余 ${times} 次`,
          icon: 'none'
        })
      })
      clearTimeout(timer)
      this.setState({
        isStart: false
      })
      return true
    } else {
      return false
    }
  }

  // 根据打乱拼图步数，计算完成游戏最低时间
  step2Time(steps) {
    const time = (steps - 8 <= 0 ? 0 : steps - 8) * 2.5 + (this.state.puzzleLevel ? 90 : 20) + ''
    return time
  }

  // 初始化数据
  init() {
    const { puzzleLevel } = this.state // 游戏难度
    const isEasy = puzzleLevel === 0
    const wid = 600 // 拼图整个宽度
    const rowNum = isEasy ? 3 : 4 // 拼图一行 个数
    const rowWid = Math.floor((wid - 30) / rowNum) // 拼图单个方块的宽度

    const neighbWhiteBlock = isEasy ? { // 空白目标隔壁的目标序号（可以移动到当前空白目标）
      0: [1, 3],
      1: [0, 2, 4],
      2: [1, 5],
      3: [0, 4, 6],
      4: [1, 3, 5, 7],
      5: [2, 4, 8],
      6: [3, 7],
      7: [4, 6, 8],
      8: [5, 7]
    } : { // 空白目标隔壁的目标序号（可以移动到当前空白目标）4*4
      0: [1, 4],
      1: [0, 2, 5],
      2: [1, 3, 6],
      3: [2, 7],
      4: [0, 5, 8],
      5: [1, 4, 6, 9],
      6: [2, 5, 7, 10],
      7: [3, 6, 11],
      8: [4, 9, 12],
      9: [5, 8, 10, 13],
      10: [6, 9, 11, 14],
      11: [7, 10, 15],
      12: [8, 13],
      13: [9, 12, 14],
      14: [10, 13, 15],
      15: [11, 14]
    }
    this.setState({
      initData: {
        isEasy,
        wid,
        rowWid,
        rowNum,
        neighbWhiteBlock
      },
      restTime: isEasy ? 8 : 20
    }, () => {
      const total = rowNum * rowNum
      const chooseIndex = total - 1
      const blocks = new Array(total).fill({}).map((item, index) => {
        const blockItem = {}
        const l = -rowWid * (index % rowNum) + 'rpx '
        const r = -Math.floor(index / rowNum) * rowWid + 'rpx'
        blockItem.position = l + r
        if (chooseIndex === index) {
          blockItem.isNone = true
        }
        blockItem.idx = index
        return blockItem
      })
      this.setState({
        blocks,
        constanceBlocks: blocks,
        chooseIndex
      })
    })
  }

  // 初始化 拼图顺序 -> 乱序拼图 -> 简易算法
  handleChange() {
    return new Promise((rs) => {
      const { chooseIndex, constanceBlocks, puzzleStep, initData } = this.state
      const { rowNum, neighbWhiteBlock } = initData
      const blocks = JSON.parse(JSON.stringify(constanceBlocks)) // 每次乱序前拼图都是完整的
      const total = rowNum * rowNum

      const newBlocks = new Array(total).fill(0) // 打乱成后的拼图
      const resultBlocks = new Array(total).fill(0).map((item, idex) => idex) // 0,1,2...,8
      /**
       * @property { Object } neighb 空白周边数组
       * @property { number[] } res 序列数组
       * @property { number } wIdx 空白所在序号
       * @property { number } stepNums 打乱顺序所走的步数
       * @property { boolean } repeat 是否重复
       * @property { number[] } oldStepNums 空白走过的序号， 判断下一步是否走的步数重复
       * @return { res, stepNums, wIdx, oldStepNums } // 返回打乱顺序后的序号， 打乱顺序的步数， 空白最终在的序号, 空白走的每步序号
       * */
      function helper(neighb, res, wIdx, stepNums, repeat, oldStepNums) {
        if (stepNums === puzzleStep) return { res, stepNums, wIdx, oldStepNums }
        const steps = neighb[wIdx]
        // 当前白名单步数
        const repIdx = steps.indexOf(repeat)
        if (repIdx > -1) {
          steps.splice(repIdx, 1)
        }
        const len = steps.length
        let step = steps[Math.floor(Math.random() * len)]
        if (oldStepNums.includes(step)) {
          step = steps[Math.floor(Math.random() * len)]
        }
        oldStepNums.push(step)
        // 更换 result 中的数字序号
        const temp = res[wIdx]
        res[wIdx] = res[step]
        res[step] = temp
        return helper(neighb, res, step, stepNums + 1, wIdx, oldStepNums)
      }
      const neighbTemp = JSON.parse(JSON.stringify(neighbWhiteBlock))
      const { res, stepNums, wIdx, oldStepNums } = helper(neighbTemp, resultBlocks, chooseIndex, 0, -1, [])

      // 计算时间范围
      const currentTime = this.step2Time(stepNums)
      console.log(oldStepNums, 'oldStepNums', blocks)
      newBlocks.forEach((item, index) => {
        newBlocks[index] = blocks[res[index]]
      })

      this.setState({
        blocks: newBlocks,
        isStart: false,
        nowSteps: 0,
        puzzleStep: stepNums,
        currentTime: currentTime,
        whiteBlockIdex: wIdx
      }, () => {
        rs()
      })
    })
  }

  // 开始游戏
  start() {
    const { isStart } = this.state
    if (isStart) return
    this.handleChange().then(() => {
      console.log(this.state.blocks, 'start')
      const timer = setInterval(() => {
        // 实时 情况
        const { currentTime, blocks } = this.state
        if (Number(currentTime) <= 0) {
          this.isRecover(blocks, true) // 时间结束验证是否通关
          clearTimeout(timer)
        }
        this.setState({ // 转换秒、毫秒
          currentTime: (currentTime - 80 / 1000) <= 0 ? 0 : (currentTime - 80 / 1000).toFixed(2) + ''
        })
      }, 80)
      this.setState({
        firstStart: false,
        isStart: true,
        timer
      })
    })
  }

  // 通过 / 游戏失败 请求
  postPuzzle(status) {
    return new Promise((rs, rj) => {
      postPuzzles(this.props.api, {
        phoneNumber: this.props.phoneNumber, // 模拟数据
        userInfo: this.props.userInfo, // 模拟数据
        status // success、 fail
      }).then(res => {
        rs(res.data && res.data.remain_times)
      }).catch((err) => { rj(err) })
    })
  }

  // 点击9个方块
  handleClick(index, isNone) {
    const { blocks, nowSteps, isStart, currentTime, whiteBlockIdex } = this.state
    // 游戏时间为0 不允许操作； 开始
    if (currentTime === '0' || !isStart) return

    // 当前是空白， 当前不可拖拽
    if (isNone || !this.isNeighbour(index)) return

    // 此时又选择了空白， 且为邻居， 则交换 （由拼图方块移动到空白方块）
    const temp = blocks[whiteBlockIdex]
    blocks[whiteBlockIdex] = blocks[index]
    blocks[index] = temp

    this.setState({
      blocks,
      whiteBlockIdex: index,
      nowSteps: nowSteps + 1
    })
    // 验证是否通过
    this.isRecover(blocks)
  }

  render() {
    const { navigationBarStyle, blocks, nowSteps, puzzleStep, currentTime, restTime, initData, firstStart } = this.state
    const { rowWid, isEasy, wid } = initData
    let front = currentTime ? currentTime.split('.')[0] : '0'
    const end = currentTime ? currentTime.split('.')[1] : '00'
    front = front - 10 < 0 ? '0' + front : front
    const time = this.step2Time(puzzleStep)
    // 留出空隙
    const itemWid = rowWid - (isEasy ? 5 : 8) + 'rpx'

    const blockItemStyle = (item) => ({
      backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/puz.png)`,
      'width': itemWid,
      'height': itemWid,
      'backgroundSize': wid + 'rpx ' + wid + 'rpx',
      'opacity': item.isNone ? '.35' : '1',
      'backgroundPosition': item.position
    })

    const styles = { ...navigationBarStyle, color: '#fff' }

    // 开始按钮背景图
    return (
      <View className='puzzle' style={{backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/puzzle_bg.png)`}}>
        <View className='puzzle-navigation-title' style={styles}>
          <Image src={image} mode='heightFix' onClick={Taro.navigateBack.bind(this)} />
          <Text style='font-size:20px;'>拼图游戏</Text>
        </View>
        <View className='puzzle-container'>
          {/* 标题 */}
          <View className='puzzle-container-title' style={{backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/part1.png)`}}>
            <View className='l'>
              <View>
                <Text>倒计时：</Text>
              </View>
              <View className='time-box'>
                <Text className='time' style={{backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/square.png)`}}>{front}</Text>
                <Text style='margin:0 5px'> : </Text>
                <Text className='time' style={{backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/square.png)`}}>{end}</Text>
              </View>
            </View>
            <View className='r'>
              <View>
                <Text>步数：</Text>
              </View>
              <View className='time-box'>
                <Text className='time' style={{backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/square.png)`}}>{nowSteps}</Text>
              </View>
            </View>
          </View>
          <View className='puzzle-container-icon'>
            <Image src={`${baseUrl}asstes/weapp/widgets/puzzle/puzzle.png`} />
          </View>
          {/* 拼图 */}
          <View className='puzzle-container-block' style={{ backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/part2.png)`, 'width': wid + 'rpx', 'height': wid + 'rpx' }}>
            <View className='bg'>
              {
                blocks.map((item, index) => {
                  return (
                    <View
                      className='puzzle-item'
                      style={blockItemStyle(item)}
                      key={index}
                      onClick={this.handleClick.bind(this, index, item.isNone)}
                    />
                  )
                })
              }
            </View>

            { firstStart && (
              <View className='popup' style={{right: isEasy ? '320rpx' : '246rpx'}}>
                <Image className='circle' src={`${baseUrl}asstes/weapp/widgets/puzzle/circle.png`} />
                <Image className='finger' src={`${baseUrl}asstes/weapp/widgets/puzzle/hand.png`} />
              </View>
            ) }

          </View>

          <View className='puzzle-container-start' style={{'backgroundImage': firstStart ? `url(${baseUrl}asstes/weapp/widgets/puzzle/btn.png)` : `url(${baseUrl}asstes/weapp/widgets/puzzle/btn-re.png)`}}>
            <View onClick={this.start.bind(this)}></View>
          </View>

          <View className='puzzle-container-introduction' style={{backgroundImage: `url(${baseUrl}asstes/weapp/widgets/puzzle/part3.png)`}}>

            <View className='puzzle-container-introduction-content'>
              {
                gameIntroduction(puzzleStep, time, restTime).map(item => {
                  return (
                    <View className='item' key={item.id}>
                      <View className='item-tit'><Text>{item.title}</Text></View>
                      { item.content.map((cont, index) => {
                        return (
                          <View className='item-cont' key={index}>
                            <Text>{cont}</Text>
                          </View>
                        )
                      }) }
                    </View>
                  )
                })
              }

            </View>

          </View>

        </View>
        {/* </ScrollView> */}
      </View>
    )
  }
}
